#!/bin/sh

# I tried: `REP_LOAD_PATH=... exec ...' but solaris /bin/sh won't play along

REP_LOAD_PATH="/usr/local/share/sawfish/lisp"
export REP_LOAD_PATH

exec rep "$0" "$@"

!#

;; main.jl -- entry point of Sawfish configurator
;;
;; Copyright (C) 2000 John Harper <john@dcs.warwick.ac.uk>
;;
;; This file is part of sawfish.
;;
;; sawfish is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.
;;
;; sawfish is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with sawfish; see the file COPYING.  If not, write to
;; the Free Software Foundation, 51 Franklin Street, Fifth Floor,
;; Boston, MA 02110-1301 USA. 

(structure ()

    (open rep
	  rep.system
	  rep.io.files
	  rep.regexp)

  (when (getenv "SAWFISH_USER_LISP_DIR")
    (setq load-path
	  (append
	   (mapcar canonical-file-name
		   (string-split ":" (getenv "SAWFISH_USER_LISP_DIR")))
	   load-path)))

  ;; This is put here *after* SAWFISH_USER_LISP_DIR is parsed.
  (require 'sawfish.cfg.shell)

  (run-shell))

;; Emacs support
;; Local variables:
;; mode: sawfish
;; End:
