;; custom-defaults.jl -- loaded if there's no user customization

;; Copyright (C) 2000 John Harper <john@dcs.warwick.ac.uk>

;; This file is part of sawfish.

;; sawfish is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; sawfish is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with sawfish; see the file COPYING.  If not, write to
;; the Free Software Foundation, 51 Franklin Street, Fifth Floor, 
;; Boston, MA 02110-1301 USA.

;; This file is read if there's no ~/.sawfish/custom.

;; The following can be taken as an example.
;; (But it is a deprecated variable, so meaningless per se.)

;;(custom-set-typed-variable (quote workspace-geometry)
;;			     (quote (4 1 . 1))
;;                           (quote workspace-geometry))
