/* build.h -- Definitions relating to the current build
   Automatically generated by build-info; DO NOT EDIT! */

#ifndef BUILD_H
#define BUILD_H

#define SAWFISH_VERSION "1.13.0"
#define SAWFISH_DIR "/usr/local/share/sawfish"
#define SAWFISH_LISPDIR "/usr/local/share/sawfish/lisp"
#define SAWFISH_EXECDIR "/usr/local/lib/sawfish"
#define SAWFISH_LOCALEDIR "/usr/local/share/locale"

#endif /* BUILD_H */
